@@ -21,22 +21,22 @@ OTHER DEALINGS IN THE SOFTWARE.
 """
 from __future__ import absolute_import
 
-import json
 import sys
 from collections import OrderedDict, namedtuple
+from distutils.util import strtobool
 from functools import partial
 from itertools import chain
 from types import ModuleType
 from wsgiref.simple_server import make_server
 
 import falcon
-from falcon import HTTP_METHODS
-
 import hug.defaults
 import hug.output_format
+from falcon import HTTP_METHODS
+from hug import introspect
+from hug._async import asyncio, ensure_future
 from hug._version import current
 
-
 INTRO = """
 /#######################################################################\\
           `.----``..-------..``.----.
@@ -75,7 +75,7 @@ class InterfaceAPI(object):
 class HTTPInterfaceAPI(InterfaceAPI):
     """Defines the HTTP interface specific API"""
     __slots__ = ('routes', 'versions', 'base_url', '_output_format', '_input_format', 'versioned', '_middleware',
-                 '_not_found_handlers', '_startup_handlers', 'sinks', '_not_found', '_exception_handlers')
+                 '_not_found_handlers', 'sinks', '_not_found', '_exception_handlers')
 
     def __init__(self, api, base_url=''):
         super().__init__(api)
@@ -98,6 +98,23 @@ class HTTPInterfaceAPI(InterfaceAPI):
         """Returns the active not found handler"""
         return getattr(self, '_not_found', self.base_404)
 
+    def urls(self):
+        """Returns a generator of all URLs attached to this API"""
+        for base_url, mapping in self.routes.items():
+            for url, _ in mapping.items():
+                yield base_url + url
+
+    def handlers(self):
+        """Returns all registered handlers attached to this API"""
+        used = []
+        for base_url, mapping in self.routes.items():
+            for url, methods in mapping.items():
+                for method, versions in methods.items():
+                    for version, handler in versions.items():
+                        if not handler in used:
+                            used.append(handler)
+                            yield handler
+
     def input_format(self, content_type):
         """Returns the set input_format handler for the given content_type"""
         return getattr(self, '_input_format', {}).get(content_type, hug.defaults.input_format.get(content_type, None))
@@ -159,14 +176,12 @@ class HTTPInterfaceAPI(InterfaceAPI):
         for middleware in (http_api.middleware or ()):
             self.add_middleware(middleware)
 
-        for startup_handler in (http_api.startup_handlers or ()):
-            self.add_startup_handler(startup_handler)
-
-        for version, handler in getattr(self, '_exception_handlers', {}).items():
-            for exception_type, exception_handler in handler.items():
-                target_exception_handlers = http_api.exception_handlers(version) or {}
-                if exception_type not in target_exception_handlers:
-                    http_api.add_exception_handler(exception_type, exception_handler, version)
+        for version, handler in getattr(http_api, '_exception_handlers', {}).items():
+            for exception_type, exception_handlers in handler.items():
+                target_exception_handlers = self.exception_handlers(version) or {}
+                for exception_handler in exception_handlers:
+                    if exception_type not in target_exception_handlers:
+                        self.add_exception_handler(exception_type, exception_handler, version)
 
         for input_format, input_format_handler in getattr(http_api, '_input_format', {}).items():
             if not input_format in getattr(self, '_input_format', {}):
@@ -187,11 +202,11 @@ class HTTPInterfaceAPI(InterfaceAPI):
 
         self.not_found_handlers[version] = handler
 
-    def documentation(self, base_url=None, api_version=None):
+    def documentation(self, base_url=None, api_version=None, prefix=""):
         """Generates and returns documentation for this API endpoint"""
         documentation = OrderedDict()
         base_url = self.base_url if base_url is None else base_url
-        overview = self.api.module.__doc__
+        overview = self.api.doc
         if overview:
             documentation['overview'] = overview
 
@@ -222,23 +237,27 @@ class HTTPInterfaceAPI(InterfaceAPI):
                         for version in applies_to:
                             if api_version and version != api_version:
                                 continue
-                            doc = version_dict.setdefault(router_base_url + url, OrderedDict())
-                            doc[method] = handler.documentation(doc.get(method, None), version=version,
-                                                                base_url=router_base_url or base_url, url=url)
+                            if base_url and router_base_url != base_url:
+                                continue
+                            doc = version_dict.setdefault(url, OrderedDict())
+                            doc[method] = handler.documentation(doc.get(method, None), version=version, prefix=prefix,
+                                                                base_url=router_base_url, url=url)
 
         documentation['handlers'] = version_dict
         return documentation
 
-    def serve(self, port=8000, no_documentation=False):
+    def serve(self, host='', port=8000, no_documentation=False, display_intro=True):
         """Runs the basic hug development server against this API"""
         if no_documentation:
             api = self.server(None)
         else:
             api = self.server()
 
-        print(INTRO)
-        httpd = make_server('', port, api)
-        print("Serving on port {0}...".format(port))
+        if display_intro:
+            print(INTRO)
+
+        httpd = make_server(host, port, api)
+        print("Serving on {0}:{1}...".format(host, port))
         httpd.serve_forever()
 
     @staticmethod
@@ -277,19 +296,19 @@ class HTTPInterfaceAPI(InterfaceAPI):
         base_url = self.base_url if base_url is None else base_url
 
         def handle_404(request, response, *args, **kwargs):
-            url_prefix = self.base_url
-            if not url_prefix:
-                url_prefix = request.url[:-1]
-                if request.path and request.path != "/":
-                    url_prefix = request.url.split(request.path)[0]
+            url_prefix = request.url[:-1]
+            if request.path and request.path != "/":
+                url_prefix = request.url.split(request.path)[0]
 
             to_return = OrderedDict()
             to_return['404'] = ("The API call you tried to make was not defined. "
                                 "Here's a definition of the API to help you get going :)")
-            to_return['documentation'] = self.documentation(url_prefix, self.determine_version(request, False))
-            response.data = json.dumps(to_return, indent=4, separators=(',', ': ')).encode('utf8')
+            to_return['documentation'] = self.documentation(base_url, self.determine_version(request, False),
+                                                            prefix=url_prefix)
+            response.data = hug.output_format.json(to_return, indent=4, separators=(',', ': '))
             response.status = falcon.HTTP_NOT_FOUND
-            response.content_type = 'application/json'
+            response.content_type = 'application/json; charset=utf-8'
+        handle_404.interface = True
         return handle_404
 
     def version_router(self, request, response, api_version=None, versions={}, not_found=None, **kwargs):
@@ -308,8 +327,7 @@ class HTTPInterfaceAPI(InterfaceAPI):
         base_url = self.base_url if base_url is None else base_url
 
         not_found_handler = default_not_found
-        for startup_handler in self.startup_handlers:
-            startup_handler(self)
+        self.api._ensure_started()
         if self.not_found_handlers:
             if len(self.not_found_handlers) == 1 and None in self.not_found_handlers:
                 not_found_handler = self.not_found_handlers[None]
@@ -320,7 +338,6 @@ class HTTPInterfaceAPI(InterfaceAPI):
 
         if not_found_handler:
             falcon_api.add_sink(not_found_handler)
-            not_found_handler
             self._not_found = not_found_handler
 
         for sink_base_url, sinks in self.sinks.items():
@@ -350,46 +367,45 @@ class HTTPInterfaceAPI(InterfaceAPI):
         falcon_api.set_error_serializer(error_serializer)
         return falcon_api
 
-    @property
-    def startup_handlers(self):
-        return getattr(self, '_startup_handlers', ())
-
-    def add_startup_handler(self, handler):
-        """Adds a startup handler to the hug api"""
-        if not self.startup_handlers:
-            self._startup_handlers = []
-
-        self.startup_handlers.append(handler)
-
 HTTPInterfaceAPI.base_404.interface = True
 
 
 class CLIInterfaceAPI(InterfaceAPI):
     """Defines the CLI interface specific API"""
-    __slots__ = ('commands', )
+    __slots__ = ('commands', 'error_exit_codes',)
 
-    def __init__(self, api, version=''):
+    def __init__(self, api, version='', error_exit_codes=False):
         super().__init__(api)
         self.commands = {}
+        self.error_exit_codes = error_exit_codes
 
-    def __call__(self):
+    def __call__(self, args=None):
         """Routes to the correct command line tool"""
-        if not len(sys.argv) > 1 or not sys.argv[1] in self.commands:
+        self.api._ensure_started()
+        args = sys.argv if args is None else args
+        if not len(args) > 1 or not args[1] in self.commands:
             print(str(self))
             return sys.exit(1)
 
-        command = sys.argv.pop(1)
-        self.commands.get(command)()
+        command = args.pop(1)
+        result = self.commands.get(command)()
+
+        if self.error_exit_codes and bool(strtobool(result.decode('utf-8'))) is False:
+            sys.exit(1)
+
+    def handlers(self):
+        """Returns all registered handlers attached to this API"""
+        return self.commands.values()
 
     def __str__(self):
-        return "{0}\n\nAvailable Commands:{1}\n".format(self.api.module.__doc__ or self.api.module.__name__,
+        return "{0}\n\nAvailable Commands:{1}\n".format(self.api.doc or self.api.name,
                                                         "\n\n\t- " + "\n\t- ".join(self.commands.keys()))
 
 
 class ModuleSingleton(type):
     """Defines the module level __hug__ singleton"""
 
-    def __call__(cls, module, *args, **kwargs):
+    def __call__(cls, module=None, *args, **kwargs):
         if isinstance(module, API):
             return module
 
@@ -397,6 +413,8 @@ class ModuleSingleton(type):
             if module not in sys.modules:
                 sys.modules[module] = ModuleType(module)
             module = sys.modules[module]
+        elif module is None:
+            return super().__call__(*args, **kwargs)
 
         if not '__hug__' in module.__dict__:
             def api_auto_instantiate(*args, **kwargs):
@@ -412,10 +430,19 @@ class ModuleSingleton(type):
 
 class API(object, metaclass=ModuleSingleton):
     """Stores the information necessary to expose API calls within this module externally"""
-    __slots__ = ('module', '_directives', '_http', '_cli', '_context')
+    __slots__ = ('module', '_directives', '_http', '_cli', '_context',
+                 '_startup_handlers', 'started', 'name', 'doc', 'cli_error_exit_codes')
 
-    def __init__(self, module):
+    def __init__(self, module=None, name='', doc='', cli_error_exit_codes=False):
         self.module = module
+        if module:
+            self.name = name or module.__name__ or ''
+            self.doc = doc or module.__doc__ or ''
+        else:
+            self.name = name
+            self.doc = doc
+        self.started = False
+        self.cli_error_exit_codes = cli_error_exit_codes
 
     def directives(self):
         """Returns all directives applicable to this Hug API"""
@@ -430,6 +457,13 @@ class API(object, metaclass=ModuleSingleton):
         self._directives = getattr(self, '_directives', {})
         self._directives[directive.__name__] = directive
 
+    def handlers(self):
+        """Returns all registered handlers attached to this API"""
+        if getattr(self, '_http'):
+            yield from self.http.handlers()
+        if getattr(self, '_cli'):
+            yield from self.cli.handlers()
+
     @property
     def http(self):
         if not hasattr(self, '_http'):
@@ -439,7 +473,7 @@ class API(object, metaclass=ModuleSingleton):
     @property
     def cli(self):
         if not hasattr(self, '_cli'):
-            self._cli = CLIInterfaceAPI(self)
+            self._cli = CLIInterfaceAPI(self, error_exit_codes=self.cli_error_exit_codes)
         return self._cli
 
     @property
@@ -458,6 +492,32 @@ class API(object, metaclass=ModuleSingleton):
         for directive in getattr(api, '_directives', {}).values():
             self.add_directive(directive)
 
+        for startup_handler in (api.startup_handlers or ()):
+            self.add_startup_handler(startup_handler)
+
+    def add_startup_handler(self, handler):
+        """Adds a startup handler to the hug api"""
+        if not self.startup_handlers:
+            self._startup_handlers = []
+
+        self.startup_handlers.append(handler)
+
+    def _ensure_started(self):
+        """Marks the API as started and runs all startup handlers"""
+        if not self.started:
+            async_handlers = [startup_handler for startup_handler in self.startup_handlers if
+                              introspect.is_coroutine(startup_handler)]
+            if async_handlers:
+                loop = asyncio.get_event_loop()
+                loop.run_until_complete(asyncio.gather(*[handler(self) for handler in async_handlers], loop=loop))
+            for startup_handler in self.startup_handlers:
+                if not startup_handler in async_handlers:
+                    startup_handler(self)
+
+    @property
+    def startup_handlers(self):
+        return getattr(self, '_startup_handlers', ())
+
 
 def from_object(obj):
     """Returns a Hug API instance from a given object (function, class, instance)"""
