@@ -35,32 +35,11 @@ import hug.api
 import hug.output_format
 import hug.types as types
 from hug import introspect
+from hug._async import asyncio_call
 from hug.exceptions import InvalidTypeData
 from hug.format import parse_content_type
 from hug.types import MarshmallowSchema, Multiple, OneOf, SmartBoolean, Text, text
 
-try:
-    import asyncio
-
-    if sys.version_info >= (3, 4, 4):
-        ensure_future = asyncio.ensure_future  # pragma: no cover
-    else:
-        ensure_future = asyncio.async  # pragma: no cover
-
-    def asyncio_call(function, *args, **kwargs):
-        loop = asyncio.get_event_loop()
-        if loop.is_running():
-            return function(*args, **kwargs)
-
-        function = ensure_future(function(*args, **kwargs), loop=loop)
-        loop.run_until_complete(function)
-        return function.result()
-
-except ImportError:  # pragma: no cover
-
-    def asyncio_call(*args, **kwargs):
-        raise NotImplementedError()
-
 
 class Interfaces(object):
     """Defines the per-function singleton applied to hugged functions defining common data needed by all interfaces"""
@@ -69,6 +48,7 @@ class Interfaces(object):
         self.api = hug.api.from_object(function)
         self.spec = getattr(function, 'original', function)
         self.arguments = introspect.arguments(function)
+        self.name = introspect.name(function)
         self._function = function
 
         self.is_coroutine = introspect.is_coroutine(self.spec)
@@ -86,7 +66,8 @@ class Interfaces(object):
         self.parameters = tuple(self.parameters)
         self.defaults = dict(zip(reversed(self.parameters), reversed(self.spec.__defaults__ or ())))
         self.required = self.parameters[:-(len(self.spec.__defaults__ or ())) or None]
-        if introspect.is_method(self.spec) or introspect.is_method(function):
+        self.is_method = introspect.is_method(self.spec) or introspect.is_method(function)
+        if self.is_method:
             self.required = self.required[1:]
             self.parameters = self.parameters[1:]
 
@@ -104,7 +85,9 @@ class Interfaces(object):
                 self.directives[name] = transformer
                 continue
 
-            if hasattr(transformer, 'load'):
+            if hasattr(transformer, 'from_string'):
+                transformer = transformer.from_string
+            elif hasattr(transformer, 'load'):
                 transformer = MarshmallowSchema(transformer)
             elif hasattr(transformer, 'deserialize'):
                 transformer = transformer.deserialize
@@ -127,7 +110,7 @@ class Interface(object):
     """
     __slots__ = ('interface', '_api', 'defaults', 'parameters', 'required', '_outputs', 'on_invalid', 'requires',
                  'validate_function', 'transform', 'examples', 'output_doc', 'wrapped', 'directives', 'all_parameters',
-                 'raise_on_invalid', 'invalid_outputs')
+                 'raise_on_invalid', 'invalid_outputs', 'map_params', 'input_transformations')
 
     def __init__(self, route, function):
         if route.get('api', None):
@@ -155,6 +138,26 @@ class Interface(object):
             self.all_parameters = set(route['parameters'])
             self.required = tuple([parameter for parameter in self.parameters if parameter not in self.defaults])
 
+        if 'map_params' in route:
+            self.map_params = route['map_params']
+            for interface_name, internal_name in self.map_params.items():
+                if internal_name in self.defaults:
+                    self.defaults[interface_name] = self.defaults.pop(internal_name)
+                if internal_name in self.parameters:
+                    self.parameters = [interface_name if param == internal_name else param for param in self.parameters]
+                if internal_name in self.all_parameters:
+                    self.all_parameters.remove(internal_name)
+                    self.all_parameters.add(interface_name)
+                if internal_name in self.required:
+                    self.required = tuple([interface_name if param == internal_name else param for
+                                           param in self.required])
+
+            reverse_mapping = {internal: interface for interface, internal in self.map_params.items()}
+            self.input_transformations = {reverse_mapping.get(name, name): transform for
+                                          name, transform in self.interface.input_transformations.items()}
+        else:
+            self.input_transformations = self.interface.input_transformations
+
         if 'output' in route:
             self.outputs = route['output']
 
@@ -195,7 +198,7 @@ class Interface(object):
     def validate(self, input_parameters):
         """Runs all set type transformers / validators against the provided input parameters and returns any errors"""
         errors = {}
-        for key, type_handler in self.interface.input_transformations.items():
+        for key, type_handler in self.input_transformations.items():
             if self.raise_on_invalid:
                 if key in input_parameters:
                     input_parameters[key] = type_handler(input_parameters[key])
@@ -211,7 +214,7 @@ class Interface(object):
                     else:
                         errors[key] = str(error)
 
-        for require in self.interface.required:
+        for require in self.required:
             if not require in input_parameters:
                 errors[require] = "Required parameter '{}' not supplied".format(require)
         if not errors and getattr(self, 'validate_function', False):
@@ -262,6 +265,17 @@ class Interface(object):
 
         return doc
 
+    def _rewrite_params(self, params):
+        for interface_name, internal_name in self.map_params.items():
+            if interface_name in params:
+                params[internal_name] = params.pop(interface_name)
+
+    @staticmethod
+    def cleanup_parameters(parameters, exception=None):
+        for parameter, directive in parameters.items():
+            if hasattr(directive, 'cleanup'):
+                directive.cleanup(exception=exception)
+
 
 class Local(Interface):
     """Defines the Interface responsible for exposing functions locally"""
@@ -316,7 +330,14 @@ class Local(Interface):
                 outputs = getattr(self, 'invalid_outputs', self.outputs)
                 return outputs(errors) if outputs else errors
 
-        result = self.interface(**kwargs)
+        if getattr(self, 'map_params', None):
+            self._rewrite_params(kwargs)
+        try:
+            result = self.interface(**kwargs)
+            self.cleanup_parameters(kwargs)
+        except Exception as exception:
+            self.cleanup_parameters(kwargs, exception=exception)
+            raise exception
         if self.transform:
             result = self.transform(result)
         return self.outputs(result) if self.outputs else result
@@ -328,12 +349,12 @@ class CLI(Interface):
     def __init__(self, route, function):
         super().__init__(route, function)
         self.interface.cli = self
+        self.reaffirm_types = {}
         use_parameters = list(self.interface.parameters)
         self.additional_options = getattr(self.interface, 'arg', getattr(self.interface, 'kwarg', False))
         if self.additional_options:
             use_parameters.append(self.additional_options)
 
-
         used_options = {'h', 'help'}
         nargs_set = self.interface.takes_args or self.interface.takes_kwargs
         self.parser = argparse.ArgumentParser(description=route.get('doc', self.interface.spec.__doc__))
@@ -371,8 +392,10 @@ class CLI(Interface):
                 if transform in (list, tuple) or isinstance(transform, types.Multiple):
                     kwargs['action'] = 'append'
                     kwargs['type'] = Text()
+                    self.reaffirm_types[option] = transform
                 elif transform == bool or isinstance(transform, type(types.boolean)):
                     kwargs['action'] = 'store_true'
+                    self.reaffirm_types[option] = transform
                 elif isinstance(transform, types.OneOf):
                     kwargs['choices'] = transform.values
             elif (option in self.interface.spec.__annotations__ and
@@ -420,23 +443,31 @@ class CLI(Interface):
 
     def __call__(self):
         """Calls the wrapped function through the lens of a CLI ran command"""
+        self.api._ensure_started()
         for requirement in self.requires:
             conclusion = requirement(request=sys.argv, module=self.api.module)
             if conclusion and conclusion is not True:
                 return self.output(conclusion)
 
+        if self.interface.is_method:
+            self.parser.prog = "%s %s" % (self.api.module.__name__, self.interface.name)
+
         known, unknown = self.parser.parse_known_args()
         pass_to_function = vars(known)
         for option, directive in self.directives.items():
             arguments = (self.defaults[option], ) if option in self.defaults else ()
             pass_to_function[option] = directive(*arguments, api=self.api, argparse=self.parser,
                                                  interface=self)
+        for field, type_handler in self.reaffirm_types.items():
+            if field in pass_to_function:
+                pass_to_function[field] = type_handler(pass_to_function[field])
 
         if getattr(self, 'validate_function', False):
             errors = self.validate_function(pass_to_function)
             if errors:
                 return self.output(errors)
 
+        args = None
         if self.additional_options:
             args = []
             for parameter in self.interface.parameters:
@@ -458,18 +489,26 @@ class CLI(Interface):
                     elif add_options_to:
                         pass_to_function[add_options_to].append(option)
 
-            result = self.interface(*args, **pass_to_function)
-        else:
-            result = self.interface(**pass_to_function)
+        if getattr(self, 'map_params', None):
+            self._rewrite_params(pass_to_function)
 
+        try:
+            if args:
+                result = self.interface(*args, **pass_to_function)
+            else:
+                result = self.interface(**pass_to_function)
+            self.cleanup_parameters(pass_to_function)
+        except Exception as exception:
+            self.cleanup_parameters(pass_to_function, exception=exception)
+            raise exception
         return self.output(result)
 
 
 class HTTP(Interface):
     """Defines the interface responsible for wrapping functions and exposing them via HTTP based on the route"""
-    __slots__ = ('_params_for_outputs', '_params_for_invalid_outputs', '_params_for_transform', 'on_invalid',
+    __slots__ = ('_params_for_outputs_state', '_params_for_invalid_outputs_state', '_params_for_transform_state',
                  '_params_for_on_invalid', 'set_status', 'response_headers', 'transform', 'input_transformations',
-                 'examples', 'wrapped', 'catch_exceptions', 'parse_body', 'private')
+                 'examples', 'wrapped', 'catch_exceptions', 'parse_body', 'private', 'on_invalid', 'inputs')
     AUTO_INCLUDE = {'request', 'response'}
 
     def __init__(self, route, function, catch_exceptions=True):
@@ -479,30 +518,44 @@ class HTTP(Interface):
         self.set_status = route.get('status', False)
         self.response_headers = tuple(route.get('response_headers', {}).items())
         self.private = 'private' in route
-
-        self._params_for_outputs = introspect.takes_arguments(self.outputs, *self.AUTO_INCLUDE)
-        self._params_for_transform = introspect.takes_arguments(self.transform, *self.AUTO_INCLUDE)
-
-        if 'output_invalid' in route:
-            self._params_for_invalid_outputs = introspect.takes_arguments(self.invalid_outputs, *self.AUTO_INCLUDE)
+        self.inputs = route.get('inputs', {})
 
         if 'on_invalid' in route:
             self._params_for_on_invalid = introspect.takes_arguments(self.on_invalid, *self.AUTO_INCLUDE)
         elif self.transform:
             self._params_for_on_invalid = self._params_for_transform
 
-        if route['versions']:
-            self.api.http.versions.update(route['versions'])
+        self.api.http.versions.update(route.get('versions', (None, )))
 
         self.interface.http = self
 
+    @property
+    def _params_for_outputs(self):
+        if not hasattr(self, '_params_for_outputs_state'):
+            self._params_for_outputs_state = introspect.takes_arguments(self.outputs, *self.AUTO_INCLUDE)
+        return self._params_for_outputs_state
+
+    @property
+    def _params_for_invalid_outputs(self):
+        if not hasattr(self, '_params_for_invalid_outputs_state'):
+            self._params_for_invalid_outputs_state = introspect.takes_arguments(self.invalid_outputs,
+                                                                                *self.AUTO_INCLUDE)
+        return self._params_for_invalid_outputs_state
+
+    @property
+    def _params_for_transform(self):
+        if not hasattr(self, '_params_for_transform_state'):
+            self._params_for_transform_state = introspect.takes_arguments(self.transform, *self.AUTO_INCLUDE)
+        return self._params_for_transform_state
+
     def gather_parameters(self, request, response, api_version=None, **input_parameters):
         """Gathers and returns all parameters that will be used for this endpoint"""
         input_parameters.update(request.params)
+
         if self.parse_body and request.content_length:
             body = request.stream
             content_type, content_params = parse_content_type(request.content_type)
-            body_formatter = body and self.api.http.input_format(content_type)
+            body_formatter = body and self.inputs.get(content_type, self.api.http.input_format(content_type))
             if body_formatter:
                 body = body_formatter(body, **content_params)
             if 'body' in self.all_parameters:
@@ -522,7 +575,6 @@ class HTTP(Interface):
             arguments = (self.defaults[parameter], ) if parameter in self.defaults else ()
             input_parameters[parameter] = directive(*arguments, response=response, request=request,
                                                     api=self.api, api_version=api_version, interface=self)
-
         return input_parameters
 
     @property
@@ -588,9 +640,11 @@ class HTTP(Interface):
         else:
             response.data = self.outputs(data, **self._arguments(self._params_for_outputs, request, response))
 
-    def call_function(self, **parameters):
+    def call_function(self, parameters):
         if not self.interface.takes_kwargs:
             parameters = {key: value for key, value in parameters.items() if key in self.all_parameters}
+        if getattr(self, 'map_params', None):
+            self._rewrite_params(parameters)
 
         return self.interface(**parameters)
 
@@ -628,12 +682,16 @@ class HTTP(Interface):
 
     def __call__(self, request, response, api_version=None, **kwargs):
         """Call the wrapped function over HTTP pulling information as needed"""
-        api_version = int(api_version) if api_version is not None else api_version
+        if isinstance(api_version, str) and api_version.isdigit():
+            api_version = int(api_version)
+        else:
+            api_version = None
         if not self.catch_exceptions:
             exception_types = ()
         else:
             exception_types = self.api.http.exception_handlers(api_version)
             exception_types = tuple(exception_types.keys()) if exception_types else ()
+        input_parameters = {}
         try:
             self.set_response_defaults(response, request)
 
@@ -648,10 +706,13 @@ class HTTP(Interface):
             if errors:
                 return self.render_errors(errors, request, response)
 
-            self.render_content(self.call_function(**input_parameters), request, response, **kwargs)
-        except falcon.HTTPNotFound:
+            self.render_content(self.call_function(input_parameters), request, response, **kwargs)
+            self.cleanup_parameters(input_parameters)
+        except falcon.HTTPNotFound as exception:
+            self.cleanup_parameters(input_parameters, exception=exception)
             return self.api.http.not_found(request, response, **kwargs)
         except exception_types as exception:
+            self.cleanup_parameters(input_parameters, exception=exception)
             handler = None
             exception_type = type(exception)
             if exception_type in exception_types:
@@ -665,11 +726,15 @@ class HTTP(Interface):
                                 handler = potential_handler
 
             if not handler:
-                raise exception_type
+                raise exception
 
             handler(request=request, response=response, exception=exception, **kwargs)
+        except Exception as exception:
+            self.cleanup_parameters(input_parameters, exception=exception)
+            raise exception
 
-    def documentation(self, add_to=None, version=None, base_url="", url=""):
+
+    def documentation(self, add_to=None, version=None, prefix="", base_url="", url=""):
         """Returns the documentation specific to an HTTP interface"""
         doc = OrderedDict() if add_to is None else add_to
 
@@ -678,7 +743,7 @@ class HTTP(Interface):
             doc['usage'] = usage
 
         for example in self.examples:
-            example_text = "{0}{1}{2}".format(base_url, '/v{0}'.format(version) if version else '', url)
+            example_text = "{0}{1}{2}{3}".format(prefix, base_url, '/v{0}'.format(version) if version else '', url)
             if isinstance(example, str):
                 example_text += "?{0}".format(example)
             doc_examples = doc.setdefault('examples', [])
@@ -724,4 +789,3 @@ class ExceptionRaised(HTTP):
         self.handle = route['exceptions']
         self.exclude = route['exclude']
         super().__init__(route, *args, **kwargs)
-
